/**
 * API to define a structured grid for the gridspec convention to libcf.
 *
 * $Id: nccf_def_grid.c 818 2011-09-13 01:22:33Z pletzer $
 */

#include "nccf_grid.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <nccf_coord.h>
#include <nccf_handle_error.h>
#include <nccf_varObj.h>

struct CFLISTITEM *CFLIST_STRUCTURED_GRID;

/** \defgroup gs_grid_grp Structured grids
    \ingroup gridspec_grp

Curvilinear, structured grids are collections of coordinates, one
for each spatial dimension. Grids keep a reference to the 
underlying coordinate objects. Therefore, coordinate objects must 
exist prior to the construction of a grid and coordinate objects
should not be freed before all operations on a structured grid have 
been completed.

*/

/**
 * \ingroup gs_grid_grp
 * Define structured grid (acts as a constructor).
 *
 * \param coordids coordinates IDs of the grid.
 * \param gridname name this grid will be given
 * \param gridid (output) grid ID
 * \return NC_NOERR on success
 *
 * \author David Kindig and Alexander Pletzer, Tech-X Corp.
 */
int nccf_def_grid(const int coordids[], const char *gridname,
                         int *gridid){
 
    /* Instantiate object */
    struct nccf_struct_grid_type *self;
    self = (struct nccf_struct_grid_type *)
             malloc(sizeof(struct nccf_struct_grid_type));

    int i, ndims, status;

    /* Initialization */
    self->coordids = NULL;
    self->ndims = 0;
    self->gridname = NULL;
    self->imask = NULL;

    /* Grid Name */
    self->gridname = (char*)calloc( STRING_SIZE, sizeof(char));
    strcpy( self->gridname, gridname );

    /* Create a new, global list of grids, if need be */
    if(CFLIST_STRUCTURED_GRID == NULL) 
            nccf_li_new(&CFLIST_STRUCTURED_GRID);

    /* Add this instance to the linked list */
    *gridid  = nccf_li_add(&CFLIST_STRUCTURED_GRID, self);

    /* Get the number of dimensions from the first coordinate 
       must have at least one element */
    if (coordids) {
      status = nccf_inq_coord_ndims(coordids[0], &ndims);
      self->ndims = ndims;

      /* Store the underlying coordinate ids */
      self->coordids = (int *) malloc(ndims * sizeof( int ));
      for(i = 0; i < ndims; i++) {
        self->coordids[i] = coordids[i];
      }
    }

    return NC_NOERR;
}
