/**
 * $Id: nccf_find_indices.h 633 2011-03-28 18:36:29Z pletzer $
 *
 * \author Alexander Pletzer, Tech-X Corp.
 */

_TYPE
nccf_get_distance_in_coord_TYPE(int ndims, const int dims[], 
				const _TYPE **coordData, 
				const _TYPE targetPos[],
				const _TYPE dindices[]) {
  int i;
  _TYPE xPos[ndims];
  nccf_get_position_TYPE(ndims, dims, coordData, dindices, xPos);
  _TYPE xNormErr = 0.0;
  _TYPE diff;
  for (i = 0; i < ndims; ++i) {
    diff = targetPos[i] - xPos[i];
    xNormErr += diff * diff;
  }
  xNormErr = sqrt(xNormErr);
  return xNormErr;
}


int
nccf_find_indices_TYPE(int ndims, const int dims[], 
		       const _TYPE **coordData, 
		       const int isPeriodic[],
		       const _TYPE targetPos[],
		       int nIterMax, _TYPE tolPos, 
		       void (*adjustFunc)(int, const int[], _TYPE []),
		       _TYPE dindices[]) {

  int status, i, iter;
  int totError = 0;

  // fix input, if need be
  nIterMax = abs(nIterMax);
  tolPos = fabs(tolPos);

  _TYPE dindices_new[ndims];
  _TYPE xPos[ndims];
  _TYPE xNormErr, xNormErrNew;
  _TYPE dIndicesBest[ndims];
  _TYPE xNormErrBest = HUGE_VAL;
  _TYPE alpha = 1.0;

  /* no tolerance should be smaller than this */
  const _TYPE eps = 0.1 * tolPos;
  const _TYPE indexTol = 0.1; // for linesearch

  for (i = 0; i < ndims; ++i) {
    dIndicesBest[i] = dindices[i];
  }

#ifdef LOGGING
    printf("-----------\ntargetPos = ");
    for (i = 0; i < ndims; ++i) printf("%f, ", targetPos[i]);
    printf("\n");
    printf("iteration alpha   xNormErr   position/indices/displ \n");
#endif

  for (iter = 0; iter < nIterMax; ++iter) {

    xNormErr = nccf_get_distance_in_coord_TYPE(ndims, dims, 
						 coordData, 
						 targetPos, 
						 dindices);

    // check if done
    if (xNormErr < tolPos) break;

    // compute the increment
    status = nccf_find_next_indices_TYPE(ndims, dims, coordData, targetPos,
					  dindices, dindices_new, xPos);
    totError += abs(status);

#ifdef LOGGING
    printf("%4d      %5.3f  %10.3g ", iter, alpha, xNormErr);
    for (i = 0; i < ndims; ++i) {
      printf("%10.6f, ", xPos[i]);
    }
    for (i = 0; i < ndims; ++i) {
      printf("%10.4f, ", dindices[i]);
    }
    for (i = 0; i < ndims; ++i) {
      printf("%10.4f, ", dindices_new[i] - dindices[i]);
    }
    printf("\n");
#endif

    // adjust indices if these fall outside the domain
    if (adjustFunc) {
      adjustFunc(ndims, dims, dindices);
    }

    // correct if overshooting
    for (i = 0; i < ndims; ++i) {
      dindices_new[i] = (dindices_new[i] < 0? 0: dindices_new[i]);
      dindices_new[i] = (dindices_new[i] > dims[i] - 1? dims[i] - 1: dindices_new[i]);
    }

    xNormErrNew = nccf_get_distance_in_coord_TYPE(ndims, dims, 
						    coordData, 
						    targetPos, 
						    dindices_new);

    // do a linesearch
    if (xNormErrNew > 0.99 * xNormErr) {
      status = nccf_linesearch_indices_TYPE(ndims, dims, coordData, 
					      targetPos, indexTol, 
					      dindices, dindices_new);
      totError += abs(status);
    }
    
    // error handling
    int bad = 0;
    for (i = 0; i < ndims; ++i) {

      if (dindices_new[i] < -eps) {
	/* a negative value means we went beyond the left boundary */
	totError = -(i + 1) * 10;
	bad = 1;
      }
      
      if (dindices_new[i] > dims[i] - 1 + eps) {
	/* a positive value means we went beyond the right boundary */
	totError = +(i + 1) * 10;
	bad = 1;
      }
    }

    if (bad) {
      // generally means we left the domain
      break;
    }

    // update
    for (i = 0; i < ndims; ++i) {
      dindices[i] += alpha*(dindices_new[i] - dindices[i]);
    }

    // are we improving?
    xNormErrNew = nccf_get_distance_in_coord_TYPE(ndims, dims, 
						    coordData, 
						    targetPos, 
						    dindices);

    // store the best result so far
    if (xNormErrNew < xNormErrBest) {
      for (i = 0; i < ndims; ++i) {
	dIndicesBest[i] = dindices[i];
      }
      xNormErrBest = xNormErrNew;
    }
    
/*     if (xNormErrNew > 0.99 * xNormErr) { */
/*       // add some relaxation */
/*       alpha *= 0.9; */
/* /\*       for (i = 0; i < ndims; ++i) { *\/ */
/* /\* 	dindices[i] += eps * (_TYPE)random()/ (_TYPE)RAND_MAX; *\/ */
/* /\*       } *\/ */
/*     } */
/*     else if (xNormErrNew < 0.7 * xNormErr) { */
/*       alpha = 1.0; */
/*     } */
  }
    
  // set the error code
  if (iter >= nIterMax) {
    totError = 1;
  }

  // make sure we return our best guess so far...
  for (i = 0; i < ndims; ++i) {
    dindices[i] = dIndicesBest[i];
  }

#ifdef LOGGING
  if (totError) {
    printf("*** Return flag: %d\n", totError);
  }
#endif
  
  return totError;
}

